import path from "node:path";
import { Glob } from "bun";
import z from "zod";
import { FileNotFoundError, MuseError } from "./errors";
import { MuseFile, type MuseFileContent } from "./muse-file";

// binding.yaml file schema
const BindingSchema = z.object({
  include: z.array(z.string()).default([]),

  name: z.string().default("Unnamed playbill"),
  author: z.string().default("Someone"),
  description: z.string().default("No description provided"),
  version: z.string().default("0.0.0"),

  files: z
    .array(z.string())
    .default(["**/*.yaml", "**/*.md", "**/*.toml", "**/*.json"]),

  options: z.record(z.string(), z.any()).default({}),
  processors: z.array(z.string()).default([]),
});

type ParsedBinding = z.infer<typeof BindingSchema>;

/**
 * Given a path, find the binding.yaml file
 * If the path is to a directory, check for a binding.yaml inside
 * If the path is to a file, check if it exists and is a binding.yaml
 * Otherwise throw an error
 *
 * @param bindingPath - The path to the binding file
 * @returns The absolute resolved path to the binding file
 */
export async function resolveBindingPath(bindingPath: string): Promise<string> {
  // If the path does not specify a filename, use binding.yaml
  const inputLocation = Bun.file(bindingPath);

  // If it is a directory, try again seeking a binding.yaml inside
  const stat = await inputLocation.stat();
  if (stat.isDirectory()) {
    return resolveBindingPath(path.resolve(bindingPath, "binding.yaml"));
  }

  // If it doesnt exist, bail
  if (!(await inputLocation.exists())) {
    throw new FileNotFoundError(bindingPath);
  }

  // Resolve the path
  return path.resolve(bindingPath);
}

export interface BindingMetadata {
  name: string;
  author: string;
  description: string;
  version: string;
}

export interface MuseProcessor {
  readonly filePath: string;
  process(binding: Binding): Promise<Binding>;
}

export interface Binding {
  readonly _raw: ParsedBinding;
  readonly bindingPath: string;
  readonly includedFiles: string[];
  readonly entries: MuseFileContent[];
  readonly metadata: BindingMetadata;
  readonly options: Record<string, unknown>;
  readonly processors: MuseProcessor[];
  readonly imports: Binding[];
}

export async function loadBinding(initialPath: string): Promise<Binding> {
  // Resolve the file location if possible
  const bindingPath = await resolveBindingPath(initialPath);
  const bindingFile = new MuseFile(bindingPath);

  // Load and parse the Binding YAML content
  const bindingFileContent = await bindingFile.read();
  const parsedBinding = BindingSchema.parse(bindingFileContent.data);

  // Prepare collections to load into
  const extendsFrom: Binding[] = [];
  const entries: MuseFileContent[] = [];

  // Process imported bindings and bring in their entries
  for (const includePath of parsedBinding.include) {
    const pathFromHere = path.resolve(bindingFile.dirname, includePath);
    const extendBindingPath = await resolveBindingPath(pathFromHere);
    const loadedBinding = await loadBinding(extendBindingPath);
    extendsFrom.push(loadedBinding);
    entries.push(...loadedBinding.entries);
  }

  // Collect file manifest from globs
  const allFilePaths: string[] = [];
  for (const thisGlob of parsedBinding.files) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingFile.dirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  // Exclude the binding.yaml file
  const includedFilePaths = allFilePaths.filter((filePath) => {
    return filePath !== bindingFile.path;
  });

  const fileLoadPromises: Promise<MuseFileContent>[] = [];
  for (const filePath of includedFilePaths) {
    const file = new MuseFile(filePath);
    fileLoadPromises.push(file.read());
  }

  const mainEntries = await Promise.all(fileLoadPromises);
  entries.push(...mainEntries);

  // Load and check processors
  const processors: MuseProcessor[] = [];
  for (const processorPath of parsedBinding.processors) {
    const resolvedProcessorPath = path.resolve(
      bindingFile.dirname,
      processorPath,
    );
    const { process } = await import(resolvedProcessorPath);

    if (!process || typeof process !== "function") {
      throw new MuseError(
        `Processor at ${processorPath} does not export a process() function`,
      );
    }

    processors.push({
      filePath: resolvedProcessorPath,
      process: process as MuseProcessor["process"],
    });
  }

  return {
    _raw: parsedBinding,
    bindingPath,
    includedFiles: includedFilePaths,
    entries,
    options: parsedBinding.options,
    processors: processors,
    imports: extendsFrom,
    metadata: {
      name: parsedBinding.name,
      author: parsedBinding.author,
      description: parsedBinding.description,
      version: parsedBinding.version,
    },
  };
}
