import chalk from "chalk";
import { MuseError } from "./errors";
import { loadBinding, type Binding } from "./binding";
import { type CLIArguments, parseCLIArguments, USAGE } from "./args";

enum ExitCode {
  Success = 0,
  Error = 1,
}

async function processBinding({ inputFilePath }: CLIArguments) {
  // Load the binding
  const binding = await loadBinding(inputFilePath);

  // Run the data through all processors
  const processedSteps: Binding[] = [binding];
  for (const processor of binding.processors) {
    const processedStep = await processor.process(binding);
    processedSteps.push(processedStep);
  }

  const finalState = processedSteps[processedSteps.length - 1];
  const serialized = JSON.stringify(finalState.entries, null, 2);

  // Otherwise
  console.log(serialized);
  return ExitCode.Success;
}

async function main(): Promise<number> {
  const cliArguments = parseCLIArguments(Bun.argv.slice(2));
  const { options } = cliArguments;

  // If --help is specified, print usage and exit
  if (options.help) {
    console.log(USAGE);
    return ExitCode.Success;
  }

  const lastProcessResult = await processBinding(cliArguments);

  return lastProcessResult;
}

try {
  const exitCode = await main();
  process.exit(exitCode);
} catch (error) {
  if (error instanceof MuseError) {
    console.error(chalk.red(error.fmt()));
  } else {
    console.error("An unexpected error occurred");
    console.error(error);
  }

  process.exit(1);
}
